//+-----------------------------------------------------------------------------+
//| Copyright (C) 2013 REGEO (http://www.regeo.uji.es/)
//|
//| This program is free software
//| You can redistribute it and/or modify it under the terms of the GNU General Public License,
//| as published by the Free Software Foundation.
//| 
//| This program is distributed in the hope that it will be useful,
//| but WITHOUT ANY WARRANTY; without even the implied warranty of
//| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//+----------------------------------------------------------------------------*/

#include <cmath>
#include <cassert>
#include <vector>

#include "Tools_Vector.h"
#include "Tools_Geometry.h"


POINT2D::POINT2D()
:x (0),
 y (0),
 merit(0)
{
}


POINT2D::POINT2D(const POINT2D& DP)
:x (DP.x),
 y (DP.y),
 merit (DP.merit)
{
}


POINT2D::POINT2D(double cx, double cy, double cmerit)
:x (cx),
 y (cy),
 merit (cmerit)
{
}


POINT2D::~POINT2D()
{
}


bool POINT2D::operator==(const POINT2D& right) const
//Overload of == operator for POINT2D
{
	return (this->x==right.x && this->y==right.y);
}


bool POINT2D::operator!=(const POINT2D& right) const
//Overload of != operator for POINT2D
{
	return (this->x!=right.x || this->y!=right.y);
}


POINT2D& POINT2D::operator+=(const POINT2D& right)
//Overload of += operator for POINT2D
{
	this->x+= right.x;
	this->y+= right.y;
	return *this;
}


POINT2D POINT2D::operator+(const POINT2D& right) const
//Overload of + operator for POINT2D
{
	POINT2D p(this->x+right.x, this->y+right.y, this->merit);
	return p;
	//return POINT2D (*this) += right;
}


POINT2D& POINT2D::operator-=(const POINT2D& right)
//Overload of -= operator for POINT2D
{
	this->x-= right.x;
	this->y-= right.y;
	return *this;
}


POINT2D POINT2D::operator-(const POINT2D& right) const
//Overload of - operator for POINT2D
{
	POINT2D p(this->x-right.x, this->y-right.y, this->merit);
	return p;
	//return POINT2D (*this) -= right;
}


POINT2D& POINT2D::operator*=(const POINT2D& right)
//Overload of *= operator for POINT2D
{
	this->x*= right.x;
	this->y*= right.y;
	return *this;
}


POINT2D POINT2D::operator*(const POINT2D& right) const
//Overload of * operator for POINT2D
{
	POINT2D p(this->x*right.x, this->y*right.y, this->merit);
	return p;
	//return POINT2D (*this) *= right;
}


POINT2D& POINT2D::operator/=(const POINT2D& right)
//Overload of /= operator for POINT2D
{
	assert(right.x!=0);
	assert(right.y!=0);
	this->x/= right.x;
	this->y/= right.y;
	return *this;
}


POINT2D POINT2D::operator/(const POINT2D& right) const
//Overload of / operator for POINT2D
{
	assert(right.x!=0);
	assert(right.y!=0);
	POINT2D p(this->x/right.x, this->y/right.y, this->merit);
	return p;
	//return POINT2D (*this) /= right;
}


void POINT2D::assign(double x_new, double y_new)
//Assigns the new coordinates to the point
{
	x= x_new;
	y= y_new;
}


POINT2D POINT2D::Rotate(POINT2D Center, double beta)
//Returns coordinates of the point rotated and angle "beta" around the point "center"
{
	POINT2D PRot;
	PRot.x= Center.x+((x-Center.x)*cos(beta)-(y-Center.y)*sin(beta));
	PRot.y= Center.y+((x-Center.x)*sin(beta)+(y-Center.y)*cos(beta));
	return PRot;
}


POINT2D POINT2D::Rotate(double beta)
//Returns coordinates of the point rotated and angle "beta" around the origin
{
	POINT2D PRot;
	PRot.x= x*cos(beta)-y*sin(beta);
	PRot.y= x*sin(beta)+y*cos(beta);
	return PRot;
}


double GetAngle_PI(POINT2D P0, POINT2D P1)
//Gets the angle (in radians) between
//	the segment defined by P0 and P1
//	and the X axis

//Since the edge is not oriented,
//returns a value in the range [0, PI)
{
	
	double angle= atan2(P1.y-P0.y, P1.x-P0.x);
	if(angle<0) angle+= NUMBER_PI;
	double length= MODULE2D(P0.x, P0.y, P1.x, P1.y);
	assert(length > 0);

	double Angle;
	if (P0.y>=P1.y)   // [0, 180]
		Angle= acos((P0.x-P1.x)/length);
	else   // (180, 360]
		Angle= acos((P1.x-P0.x)/length);

	if (fabs(Angle-NUMBER_PI)<Accuracy) Angle= 0.0;

	return Angle;
}


double GetDistancePointToSegment(POINT2D p0, POINT2D p1, POINT2D V,
								 POINT2D &Intersection)
//Gets the distance from an external point "V" to a segment defined by p0 and p1
//Gets also the intersection of p0-p1 with the minimum distance segment

//Returns the distance and the intersection point
{
	if(MODULE_2D(p0, p1) == 0)
		return MODULE_2D(p0, V);

	//Get the intersection point
	double u= (V.x - p0.x)*(p1.x - p0.x) + (V.y - p0.y)*(p1.y - p0.y);
	u/= ((p1.x - p0.x)*(p1.x - p0.x) + (p1.y - p0.y)*(p1.y - p0.y));
	Intersection.x = p0.x + ((p1.x - p0.x)*u);
	Intersection.y = p0.y + ((p1.y - p0.y)*u);

	//Get the distance from the external point to the intersection point
	return MODULE_2D(Intersection, V);
}


double GetDistancePointToSegment(POINT2D p0, POINT2D p1, POINT2D V)
//Gets the distance from an external point "V" to a segment defined by p0 and p1
{
	double mod= MODULE_2D(p0, p1);
	if(mod == 0)
		return MODULE_2D(p0, V);

	return fabs(((V.x - p0.x)*(p1.y - p0.y) - (V.y - p0.y)*(p1.x - p0.x)) / mod);
}


bool AreCollinear(POINT2D A0, POINT2D A1,
				  POINT2D B0, POINT2D B1,
				  double ThresholdAngle, double ThresholdDistance)
//Returns true if two segments are collinear, withing a "treshold"
//A0 and A1 are the endpoints of the first segment
//B0 and B1 are the endpoints of the second segment

//The ThresholdAngle is the maximum angle difference (in radians)

//The ThresholdDistance is the maximum perpendicular distance
//between the end nodes of the second segment
//and the line defined by the first segment
//(in % of the minimum lenght of the segments)
{
	double Offset= std::min(MODULE_2D(A0,A1), MODULE_2D(B0,B1));
	Offset*= ThresholdDistance; //% of the legth of the shortest segment

	//Check if segments are nearly coincident
	//---------------------------------------
	if((MODULE_2D(A0,B0)<Offset && MODULE_2D(A1,B1)<Offset) ||
	   (MODULE_2D(A0,B1)<Offset && MODULE_2D(A1,B0)<Offset))
		return true; //Lines overlap (are trivially collinear)

	//Check if worst point of one line is too distant from the other
	//--------------------------------------------------------------
	if(GetDistancePointToSegment(A0, A1, B0) > Offset &&
	   GetDistancePointToSegment(A0, A1, B1) > Offset)
	   return false;

	//Check orientation
	//-----------------
	double angleA= GetAngle_PI(A0,A1);
	double angleB= GetAngle_PI(B0,B1);

	//Return true if differences between both angles are below the threshold
	return ((fabs(angleA-angleB) < ThresholdAngle ||
			 fabs(fabs(angleA-angleB)-NUMBER_PI) < ThresholdAngle));

	////Get the angles of the line that connects
	////the two most distant points (c0,c1) and the two closest points (d0,d1)
	//POINT2D c0, d0;
	//POINT2D bM((B0.x+B1.x)/2, (B0.y+B1.y)/2);
	//if(MODULE_2D(A0,bM) > MODULE_2D(A1,bM)){
	//	c0= A0;
	//	d0= A1;
	//}
	//else{
	//	c0= A1;
	//	d0= A0;
	//}

	//POINT2D c1, d1;
	//POINT2D aM((A0.x+A1.x)/2, (A0.y+A1.y)/2);
	//if(MODULE_2D(aM,B0) > MODULE_2D(aM,B1)){
	//	c1= B0;
	//	d1= B1;
	//}
	//else{
	//	c1= B1;
	//	d1= B0;
	//}

	//double angleC= GetAngle_PI(c0, c1);
	//double angleD= GetAngle_PI(d0, d1);

	////Return true if differences between all four angles are below the threshold
	//return ((fabs(angleA-angleB) < ThresholdAngle ||
	//		 fabs(fabs(angleA-angleB)-NUMBER_PI) < ThresholdAngle) &&
	//		 (fabs(angleA-angleC) < ThresholdAngle ||
	//		 fabs(fabs(angleA-angleC)-NUMBER_PI) < ThresholdAngle) &&
	//		 (fabs(angleA-angleD) < ThresholdAngle ||
	//		 fabs(fabs(angleA-angleD)-NUMBER_PI) < ThresholdAngle));
}


bool AreCollinear(double a0x, double a0y,
				  double a1x, double a1y,
				  double b0x, double b0y,
				  double b1x, double b1y,
				  double ThresholdAngle, double ThresholdDistance)
//Returns true if two segments are collinear, withing a "treshold"
//(a0x,a0y) and (a1x,a1y) are the endpoints of the first segment
//(b0x,b0y) and (b1x,b1y) are the endpoints of the second segment

//The ThresholdAngle is the maximum angle difference (in radians)

//The ThresholdDistance is the maximum distance
//between the end nodes of the second segment
//and the line defined by the first segment
//(in % of the minimum lenght of the segments)
{
	POINT2D A0(a0x, a0y, 0.);
	POINT2D A1(a1x, a1y, 0.);

	POINT2D B0(b0x, b0y, 0.);
	POINT2D B1(b1x, b1y, 0.);

	return AreCollinear(A0, A1, B0, B1,
						ThresholdAngle, ThresholdDistance);
}

double GetAngle_PI(POINT2D P1, POINT2D P2, POINT2D P3)
//Gets the angle (in radians) between the two segments
//defined by P1-P2 and P2-P3

//Returns a value in the range [0, PI]
{
	P1-= P2;
	P3-= P2;
	double norm_cross= P1.x*P3.y-P1.y*P3.x;
	double dot= P1.x*P3.x+P1.y*P3.y;
	return abs(atan2(norm_cross, dot));
}